/*************************   TIFFLOOK  ******************************
**  Purpose:      Display information in a TIFF data structure.
**  Author:       James D. Murray, Anaheim, CA  USA
**  Date:         September 1990
**  C Compilers:  ANSI
**  Validated on: MSC 6.0, Turbo C++ 1.0
**  Memory Model: Any
**
**  The TiffLook() function is called to display the information
**  found in a valid TIFFIMAGE structure.  This function also
**  gives valuable insight on how to correctly read a TIFFIMAGE
**  structure when writing code for other applications.
**
**  Copyright (C) 1990 by The C Gazette.  All rights reserved.
**  May be used freely if author and publication are acknowledged.
********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tiff.h"

#define NUMOFTAGNAMES                48 /* Number of TIFF tags */

struct {
    SHORT tagnumber;
    ASCII tagname[26];
} tagnames[NUMOFTAGNAMES] = {
    { T_NEWSUBFILETYPE,            "NewSubFileType" },
    { T_SUBFILETYPE,               "SubFileType" },
    { T_IMAGEWIDTH,                "ImageWidth" },
    { T_IMAGELENGTH,               "ImageLength" },
    { T_BITSPERSAMPLE,             "BitsPerSample" },
    { T_COMPRESSION,               "Compression" },
    { T_PHOTOMETRICINTERPRETATION, "PhotometricInterpretation" },
    { T_THRESHHOLDING,             "Threshholding" },
    { T_CELLWIDTH,                 "CellWidth" },
    { T_CELLLENGTH,                "CellLength" },
    { T_FILLORDER,                 "FillOrder" },
    { T_DOCUMENTNAME,              "DocumentName" },
    { T_IMAGEDESCRIPTION,          "ImageDescription" },
    { T_MAKE,                      "Make" },
    { T_MODEL,                     "Model" },
    { T_STRIPOFFSETS,              "StripOffsets" },
    { T_ORIENTATION,               "Orientation" },
    { T_SAMPLESPERPIXEL,           "SamplesPerPixel" },
    { T_ROWSPERSTRIP,              "RowsPerStrip" },
    { T_STRIPBYTECOUNTS,           "StripByteCounts" },
    { T_MINSAMPLEVALUE,            "MinSampleValue" },
    { T_MAXSAMPLEVALUE,            "MaxSampleValue" },
    { T_XRESOLUTION,               "XResolution" },
    { T_YRESOLUTION,               "YResolution" },
    { T_PLANARCONFIGURATION,       "PlanarConfiguration" },
    { T_PAGENAME ,                 "PageName" },
    { T_XPOSITION,                 "XPosition" },
    { T_YPOSITION,                 "YPosition" },
    { T_FREEOFFSETS,               "FreeOffsets" },
    { T_FREEBYTECOUNTS,            "FreeByteCounts" },
    { T_GRAYRESPONSEUNIT,          "GrayResponseUnit" },
    { T_GRAYRESPONSECURVE,         "GrayResponseCurves" },
    { T_GROUP3OPTIONS,             "Group3Options" },
    { T_GROUP4OPTIONS,             "Group4Options" },
    { T_RESOLUTIONUNIT,            "ResolutionUnit" },
    { T_PAGENUMBER,                "PageNumber" },
    { T_COLORRESPONSECURVES,       "ColorResponseCurves" },
    { T_SOFTWARE,                  "Software" },
    { T_DATETIME,                  "DateTime" },
    { T_ARTIST,                    "Artist" },
    { T_HOSTCOMPUTER,              "HostComputer" },
    { T_PREDICTOR,                 "Predictor" },
    { T_WHITEPOINT,                "WhitePoint" },
    { T_PRIMARYCHROMATICITIES,     "PrimaryChromaticities" },
    { T_COLORMAP,                  "ColorMap" },
    { T_BADFAXLINES,               "BadFaxLines" },
    { T_CLEANFAXDATA,              "CleanFaxData" },
    { T_CONSECUTIVEBADFAXLINES,    "ConsecutiveBadFaxlines" }
};

extern short (*fgetshort)(FILE *);
extern long  (*fgetlong)(FILE *);

/*
**  Look at the contents of a TIFFIMAGE structure.
**
**  This function is a parser that reads the information contained
**  within a TIFFIMAGE structure and writes it to the standard output.
**
**  Parameters: image - a valid pointer to a TIFFIMAGE structure.
**
**  Returns: Nothing.
*/
void TiffLook(TIFFIMAGE *image)
{
    LONG      i, j, k;  /* Counters */
    short     s;        /* Counter */
    long      offset;   /* Offset value holder */
    BYTE     *t_byte;   /* Index variable for byte-oriented data */
    ASCII    *t_ascii;  /* Index variable for ascii-oriented data */
    SHORT    *t_short;  /* Index variable for short-oriented data */
    LONG     *t_long;   /* Index variable for long-oriented data */
    RATIONAL *num;      /* Index variable for rational-oriented data */
    RATIONAL *den;      /* Index variable for rational-oriented data */
    TIFFIFD  *ifd;      /* IFD structure */
    TIFFTAG  *tag;      /* Tag structure */

    printf("\nTIFF File Name:   %s\n", image->TIFF_FileName);
    printf("Number of Images: %d\n", image->TIFF_NumOfImages);
    printf("Byte format:      %s\n",
      image->Header->ByteOrder == TIFF_MOTOROLA ? "Motorola" : "Intel");
    printf("Version Number:   %d\n", image->Header->Version);

    offset = image->Header->IFDOffset;
    ifd = image->IfdList;
    for (i = 1UL; i <= image->TIFF_NumOfImages; i++) {
        printf("\nImage %lu\n", i);
        printf("Offset to IFD:  0x%04lx\n", offset);
        printf("Number of Tags: %d\n", ifd->NumDirEntries);
        printf
        ("\nNumber  ID    Name               Type      Length     Data\n");

        tag = ifd->TagList;
        for (j = 1UL; j <= ifd->NumDirEntries; j++) {
            printf("  %2lu  %5u   %-19.18s%-10s%6lu", j, tag->TagID,
              TagIdString(tag->TagID), TagTypeString(tag->TagType),
              tag->TagLength);
            if (!tag->Value) {  /* Tag data is located within tag->Value */
                printf("  %-8s", "Value");
                switch (tag->TagType) {
                    case TYPE_BYTE:
                        for (k = 0UL; k < tag->TagLength; k++)
                            printf("%c", *(BYTE *) tag->ValueOffset++);
                        putchar('\n');
                        break;
                    case TYPE_ASCII:
                        for (k = 0UL, s = 0; k < tag->TagLength; k++, s += 8)
                            printf("%c", 
                                  (tag->ValueOffset >> s) & 0x000000ff);
                        putchar('\n');
                        break;
                    case TYPE_SHORT:
                        if (image->Header->ByteOrder == TIFF_INTEL) {
                            printf("%u", 
                                    tag->ValueOffset & 0x0000ffff);
                            if (tag->TagLength > 1UL)
                                printf(" %u\n", 
                                      (tag->ValueOffset >> 16) & 0x0000ffff);
                            else
                                putchar('\n');
                        } else
                        if (image->Header->ByteOrder == TIFF_MOTOROLA) {
                            printf("%u", 
                                  (tag->ValueOffset >> 16) & 0x0000ffff);
                            if (tag->TagLength > 1UL)
                                printf(" %u\n", 
                                       tag->ValueOffset & 0x0000ffff);
                            else
                                putchar('\n');
                        } else
                            putchar('\n');
                        break;
                    case TYPE_LONG:
                        printf("%lu\n", tag->ValueOffset);
                        break;
                    default:
                        break;
                }
            } else {    /* Tag data is located at an offset value */
                printf("  %-8s0x%04lx\n\t", "Offset", tag->ValueOffset);
                switch (tag->TagType) {
                    case TYPE_BYTE:
                        t_byte = (BYTE *) tag->Value;
                        for (k = 0UL; k < tag->TagLength; k++)
                            printf("%c ", *t_byte++);
                        putchar('\n');
                        break;
                    case TYPE_ASCII:
                        t_ascii = (ASCII *) tag->Value;
                        for (k = 0UL; k < tag->TagLength; k++)
                            printf("%c", *t_ascii++);
                        putchar('\n');
                        break;
                    case TYPE_SHORT:
                        t_short = (SHORT *) tag->Value;
                        for (k = 0UL; k < tag->TagLength; k++)
                            printf("%u ", *t_short++);
                        putchar('\n');
                        break;
                    case TYPE_LONG:
                        t_long = (LONG *) tag->Value;
                        for (k = 0UL; k < tag->TagLength; k++)
                            printf("%ld ", *t_long++);
                        putchar('\n');
                        break;
                    case TYPE_RATIONAL:
                    {
                        num = (RATIONAL *) tag->Value;
                        den = (RATIONAL *) tag->Value;
                        for (k = 0UL, den++; 
                             k < tag->TagLength; 
                             k++, num++, den++) {
                            printf("%ld/%ld (%f) ",
                              *num, *den, (double) (*num / *den));
                        }
                        putchar('\n');
                        break;
                    }
                    default:
                        break;
                }
            }
            tag = tag->Next;    /* Get the next tag in the linked list */
        }
        ifd = ifd->Next;        /* Get the next IFD in the linked list */
        offset = image->IfdList->NextIFDOffset;
    }
    printf("\nEnd of TIFF image file %s\n\n", image->TIFF_FileName);
}

/*
**  Match a tag data type to a tag data type string.
**
**  Parameters: tagtype - the data type of a TIFF tag.
**
**  Returns: A pointer to a string of the tag type name.
*/
char *TagTypeString(SHORT tagtype)
{
    switch (tagtype) {
        case TYPE_BYTE:
            return("BYTE");
        case TYPE_ASCII:
            return("ASCII");
        case TYPE_SHORT:
            return("SHORT");
        case TYPE_LONG:
            return("LONG");
        case TYPE_RATIONAL:
            return("RATIONAL");
        default:
            return("UNKNOWN");
    }
}

/*
**  Match a TIFF tag ID to a TIFF tag ID name.
**
**  Search the structure tagnames for a tag name using the
**  value tagid as an index.
**
**  Parameter: tagid - The ID number of a TIFF tag.
**
**  Returns: A pointer to a tag name string.
*/
char *TagIdString(SHORT tagid)
{
    register int i;

    if (tagid > 32767)      /* Reserved tags are not known */
        return("UNKNOWN");

    /*
    ** Loop through the list till we've got our tag.
    */
    for (i = 0; i < NUMOFTAGNAMES; i++) {
        if (tagnames[i].tagnumber == tagid)
            return((char *) tagnames[i].tagname);
    }
    return("UNKNOWN");      /* Did not find a matching tag number */
}